"""
Generic recursive tree reduce algorithm
=======================================

Trees are one of the most ubiquitous data structures. It is amazing how often we 
as programmers tend to reimplement the same algorithms for different trees.

This module defines a generic tree-reduce algorithms that can be
used with any tree-like object such as filesystem paths, lists, nested 
dictionaries an expression tree or even specialized tree classes! The only thing 
that must be provided is a function to get child nodes from a parent node.

Examples:

##  Use with any kind of structured tree!
```
    tree = {
        'label':'A', 'children':[
            {'label':'B', 'children':[]},
            {'label':'C', 'children': [
                {'label':'D', 'children':[]}, 
                {'label':'E', 'children':[]}
            ]}
        ]
    }

    def is_leaf(node): return isinstance(node, str)

    def children(node):
        return node['children']

    def all_nodes(node, children):
        nodes = set()
        for child in children:
            if is_leaf(child):
                nodes.add(child)
            else:
                nodes.update(child)
        return nodes | {node['label']}
        
    tree_reduce(tree, children, all_nodes)

    # Output --> {'A','B','C','D','E'}
```

##  Even with user-defined classes!
```
    class Tree:
        def __init__(self, label, children=None):
            self.label = label
            self.children = children if children else []

        def is_leaf(self):
            return len(self.children) == 0

    tree = Tree('A', [
            Tree('B'),
            Tree('C',[Tree('D'),Tree('E')])
        ]
    )

    def get_children(node):
        return node.children

    def node_to_newick(node, children):
        if node.is_leaf():
            return node.label
        else:
            return f"({','.join(children)})"


    tree_reduce(tree, get_children, node_to_newick)
                
    # Output --> '(B,(D,E))'
```

## Compose to perform complex algorithms
```
    tree = (('A',('B',('C','D'))),'E')

    def is_leaf(node): 
        return isinstance(node, str)

    def get_children(node): 
        return node if not is_leaf(node) else []

    def induced_subtree(leafs):
        def induced_subtree_generator(node, children):
            if children:
                return tuple(ch for ch in children if not ch is None)
            else:
                return node if node in leafs else None
        return induced_subtree_generator

    leafs = ['B', 'D', 'E']
    induced_subtree = tree_reduce(tree, get_children, induced_subtree(leafs))
    print(induced_subtree)

    # Output --> ((('B',('D',)),),'E')


    def merge_unary_nodes(node, children):
        if is_leaf(node):
            return node
        
        new_children = [
            ch[0] if (len(ch) == 1) else ch
            for ch in children
        ]
        
        return tuple(new_children)

    tree_reduce(induced_subtree, get_children, merge_unary_nodes)

    # Output --> (('B','D'),'E')
```
"""

from typing import (Any, List, Callable)

TreeLike = Any
Reduced = Any

def tree_reduce(
        tree_node: TreeLike, 
        get_children: Callable[[TreeLike], List[TreeLike]], 
        reduce_fn: Callable[[TreeLike, List[Reduced]], Reduced]
    ) -> Reduced:
    """
    
    Perform a recursive tree reduction for the tree-like object `tree_node` 
    taken as the root. 
    The callable `get_children` returns a list of the descendants of any node of 
    the tree.

    Example:
    ```
        tree = [[['A', 'B'], 'C'], ['D', ['E', 'F'],'G']]

        def is_leaf(node):
            return isinstance(node, str)

        def get_children(node):
            if is_leaf(node):
                return []
            else:
                return node

        def node_to_newick(node, children):
            if is_leaf(node):
                return node
            else:
                return f"({','.join(children)})"

        # <--- Actual reduction
        tree_reduce(tree, get_children, node_to_newick)
            
        # Output --> '(((A,B),C),(D,(E,F),G))'
    ```
    """

    reduced_children = [tree_reduce(child, get_children, reduce_fn) 
                        for child in get_children(tree_node)]
    return reduce_fn(tree_node, reduced_children)
# ---

"""
--------------------------------------------------------------------------------
Usage and tests
--------------------------------------------------------------------------------
To run tests: `pytest tree_reduce.py`
To run static type check: `mypy tree_reduce.py`
To run coverage analysis:
    `coverage run --src=. -m pytest tree_reduce.py`
    `coverage html`
"""
def test_tree_reduce_to_list_all_nodes():
    tree = {
        'label':'A', 'children':[
            {'label':'B', 'children':[]},
            {'label':'C', 'children': [
                {'label':'D', 'children':[]}, 
                {'label':'E', 'children':[]}
            ]}
        ]
    }

    def is_leaf(node): return isinstance(node, str)

    def children(node): return node['children']

    def all_nodes(node, children):
        nodes = set()
        for child in children:
            if is_leaf(child):
                nodes.add(child)
            else:
                nodes.update(child)
        return nodes | {node['label']}
        
    assert tree_reduce(tree, children, all_nodes) == {'A','B','C','D','E'}

def test_tree_reduce_for_newick_output():
    tree = [[['A', 'B'], 'C'], ['D', ['E', 'F'],'G']]

    children_fn = (lambda node: 
        node 
            if isinstance(node, list) 
            else []
     )

    def node_to_newick(node, children):
        print("assembling node to newick form:", node)
        return (
            '(' + ','.join(children) + ')'
            if children
            else node
        )

    assert (
        tree_reduce(tree, children_fn, node_to_newick)
        ==
        '(((A,B),C),(D,(E,F),G))'
    )


def test_tree_reduce_for_induced_subtree():
    tree = (('A',('B',('C','D'))),'E')

    children_fn = (lambda node: 
        node 
            if isinstance(node, tuple) 
            else []
     )

    def induced_subtree(leafs):
        def induced_subtree_generator(node, children):
            print('Processing node:', node)
            if children:
                return tuple(ch for ch in children if not ch is None)
            else:
                return node if node in leafs else None
        return induced_subtree_generator

    leafs = ['B', 'D', 'E']
    induced_subtree = tree_reduce(tree, children_fn, induced_subtree(leafs))
    assert( 
        induced_subtree
        ==
        ((('B',('D',)),),'E')
    )

    def merge_unary_nodes(node, children):
        is_leaf = lambda node: isinstance(node, str)
        if is_leaf(node):
            return node
        
        new_children = [
            ch[0] if (len(ch) == 1) else ch
            for ch in children
        ]
        
        return tuple(new_children)
    
    assert( 
        tree_reduce(induced_subtree, children_fn, merge_unary_nodes)
        ==
        (('B','D'),'E')
    )

"""
Ideas for property-based testing:
    - Different paths, same destination: Different traversals must yield the 
        same nodes.

    - There, and back again: Converting a tree to another format (lists - dicts)
        and back must yield the same tree.
    
    - Some things never change: Maybe check for invariants?

    - The more things change, the more they stay the same: Applying an operation 
        twice yields the same result as applying it once. Maybe filtering or 
        prunning? Traversing twice?

    - Solve a small problem first: Trees should have recursive properties given 
        their nature.

    - Hard to prove, easy to verify: Maybe yield nodes with paths and then 
        verify that following the path the node is found.
"""



"""
A simple function `parse_newick` to parse tree structures specified in the 
Newick format (NHX supported). The function allows to customize the final
representation, be it lists, dicts or a custom Tree class.

Usage
-----
```
    def tree_as_dict(label, children, distance, features):
        if children:
            return {label: children}
        else:
            return label

    parse_newick('(A,(B,C));', aggregator=tree_as_dict)

    # Output -> {'': ['A', {'': ['B', 'C']}]}
```

The function allows to customize the parsing of the raw feature and distance 
strings with the keyword-arguments `feature_parser` and `distance_parser`, for 
example, to read features in the NHX (New Hampshire eXtended Newick) format, you 
can use the following:
```
    def parse_NHX_features(feature_string):
        '''
        Example: 
          '&&NHX:conf=0.01:name=INTERNAL' -> {'conf':'0.01', 'name': 'INTERNAL'}
        '''
        prefix = '&&NHX:'
        feature_string = feature_string[len(prefix):]
        items = feature_string.split(':')
        key_value_pairs = (item.split('=') for item in items)
        return dict(key_value_pairs)

    def tree_builder(label, children, distance, features):
        return {'label': label, 'children': children, 'features': features}
    
    parse_newick(
        'A[&&NHX:conf=0.01:name=A];',
        aggregator=tree_builder,
        feature_parser=parse_NHX_features
    )

    # Output -> 
    # {'label': 'A', 'children': [], 'features': {'conf':'0.01', 'name': 'A'}}
```

To run tests: `python3 -m pytest parse_newick.py`
To type check: `python3 -m mypy parse_newick.py``
"""

from typing import (Optional, List, Any, Sequence, Tuple, TypeVar, Callable)
from functools import partial

def _find_closing(
        string: str, 
        start: int = 1, 
        pair: Sequence[str] = '()') -> int:
    """
    Find the index of the first closing parenthesis that is not matched to an 
    opening one starting from the given `start` position. 

    Example: '((),())()' -> 6
                    ^

    The `pair` argument allows to specify diferent opening/closing types
    of parenthesis.

    Example:
        _find_closing_brackets = partial(_find_closing, pair='[]')
    """
    opening = pair[0]
    closing = pair[1]

    next_closing = string.index(closing, start)

    has_next_opening = (string.find(opening, start) != -1)
    if not has_next_opening:
        return next_closing
    
    next_opening = string.index(opening, start)
    if next_closing < next_opening:
        return next_closing

    skip = _find_closing(string, next_opening+1, pair)
    return _find_closing(string, skip+1, pair)
# ---

_find_closing_parenthesis = partial(_find_closing, pair='()')
# ---

def _parts_of_subtree(newick: str) -> Tuple[str, str, str, str]:
    """
    A subtree consists of:
        children, label, branch length/support, comments/features

    Example:
        '(A,B)root:10.0[x=xx]' -> ['(A,B)', 'root', '10.0', 'x=xx']
    """

    children = ''
    if newick.startswith('('):
        # Node has children
        children_end = _find_closing_parenthesis(newick)
        children = newick[1:children_end]
        rest = newick[children_end+1:]
    else:
        rest = newick

    comment = ''
    if '[' in rest: # Extract comments from the end
        rest, comment = rest.split('[', maxsplit=1)
        comment = comment[:-1] # remove ']' from the end

    if ':' in rest: # Extract branch length
        label, length = rest.split(':', maxsplit=1)
    else:
        label = rest
        length = ''

    return children, label.strip(), length, comment
# ---

def _next_node_end(nodes_str: str) -> int:
    """
    From a comma-sepparated list of newick-formatted nodes, return the final 
    position of the first one.

    Examples: 
        '(A:1,(C[x],D))name:1.[c], (X,Y),,[xxx]' -> 23
        '(X,Y),,[xxx]' -> 5
        '[xxx]' -> 4
    """
    nodes_str = nodes_str.strip()

    current_end = 0
    # Skip children
    if nodes_str.startswith('('):
        current_end = _find_closing_parenthesis(nodes_str)
    
    # Skip label, distances and comments
    # Strategy: find the next comma that is not surrounded by brackets or
    #    parentheses or the end of the string
    grouping = {'(':'()', '[':'[]'}
    while current_end < len(nodes_str):

        char = nodes_str[current_end]

        if char in grouping:
            current_end = _find_closing(nodes_str, current_end+1, grouping[char])
            continue

        if char == ',':
            # Ah, the lonely comma...
            return current_end-1
        
        current_end += 1
    else:
        return len(nodes_str)-1
# ---

def _split_nodes(nodes_str: str) -> List[str]:
    """
    Separate the nodes from a comma-sepparated list.

    Example:  '(a,b), , :12, c[xxx]' -> ['(a,b)', '', ':12', 'c[xxx]']
    """

    nodes_str = nodes_str.strip()

    if nodes_str == '':
        return []

    if nodes_str == ',':
        return ['', '']

    if nodes_str.startswith(','):
        rest = nodes_str[1:]
        return [''] + _split_nodes(rest)

    if nodes_str.endswith(','):
        rest = nodes_str[:-1]
        return _split_nodes(rest) + ['']

    next_node_end = _next_node_end(nodes_str)
    node = nodes_str[:next_node_end+1]
    rest = nodes_str[next_node_end+1:].lstrip()

    if rest.startswith(','): 
        rest = rest[1:]

    return [node] + _split_nodes(rest)
# ---


def _simple_distance(dist: str) -> Optional[float]: 
    return float(dist) if dist else None

def _simple_feature(feat:str) -> str: 
    return feat

Distance = Any
Feature = Any
Tree = TypeVar('Tree')

def parse_newick_subtree(
        newick: str, 
        aggregator: Callable[[str, List[Tree], Distance, Feature], Tree],
        distance_parser: Callable[[str], Distance] = _simple_distance,
        feature_parser: Callable[[str], Feature] = _simple_feature) -> Tree:
    """
    Parse recursively the newick representation of a single newick node. 
    Root and child nodes are assembled using the aggregator function.

    Example: 
    ```
    def tree_as_list(label, children, distance, features):
        if not children: return label
        return children

    parse_newick_subtree('(A,(B,C))', tree_as_list)

    # Output -> ['A', ['B', 'C']]
    ```

    The distance parser and the feature parser allow to modify the parsing of
    the raw strings, for example, to read features in the NHX (extended Newick) 
    format, you can use the following:
    ```
    def parse_NHX_features(feature_string):
        '''
        Example: 
          '&&NHX:conf=0.01:name=INTERNAL' -> {'conf':'0.01', 'name': 'INTERNAL'}
        '''
        prefix = '&&NHX:'
        feature_string = feature_string[len(prefix):]
        items = feature_string.split(':')
        key_value_pairs = (item.split('=') for item in items)
        return dict(key_value_pairs)

    def tree_builder(label, children, distance, features):
        return {'label': label, 'children': children, 'features': features}
    
    parse_newick_subtree(
        'A[&&NHX:conf=0.01]',
        aggregator=tree_builder,
        feature_parser=parse_NHX_features
    )

    # Output -> 
    # {'label': 'A', 'children': [], 'features': {'conf':'0.01'}}
    """

    parts = _parts_of_subtree(newick.strip())
    (children_str, label, distance_str, comment_str) = parts
    
    children = [parse_newick_subtree(
                    child_str, 
                    aggregator, 
                    distance_parser=distance_parser, 
                    feature_parser=feature_parser) 
                for child_str in _split_nodes(children_str)]
    
    features = feature_parser(comment_str)
    distance = distance_parser(distance_str)

    return aggregator(label, children, distance, features)
# ---

def parse_newick(
        newick: str, 
        aggregator: Callable[[str, List[Tree], Distance, Feature], Tree],
        distance_parser: Callable[[str], Distance] = _simple_distance,
        feature_parser: Callable[[str], Feature] = _simple_feature) -> Tree:
    """
    Parse recursively the newick representation of a complete newick tree. 
    Root and child nodes are assembled using the aggregator function.

    Example: 
    ```
    def tree_as_list(label, children, distance, features):
        if not children: return label
        return children

    parse_newick('(A,(B,C));', tree_as_list)

    # Output -> ['A', ['B', 'C']]
    ```

    The distance parser and the feature parser allow to modify the parsing of
    the raw strings, for example, to read features in the NHX (extended Newick) 
    format, you can use the following:
    ```
    def parse_NHX_features(feature_string):
        '''
        Example: 
          '&&NHX:conf=0.01:name=INTERNAL' -> {'conf':'0.01', 'name': 'INTERNAL'}
        '''
        prefix = '&&NHX:'
        feature_string = feature_string[len(prefix):]
        items = feature_string.split(':')
        key_value_pairs = (item.split('=') for item in items)
        return dict(key_value_pairs)

    def tree_builder(label, children, distance, features):
        return {'label': label, 'children': children, 'features': features}
    
    parse_newick(
        'A[&&NHX:conf=0.01:name=A];',
        aggregator=tree_builder,
        feature_parser=parse_NHX_features
    )

    # Output -> 
    # {'label': 'A', 'children': [], 'features': {'conf':'0.01', 'name': 'A'}}
    ```
    """

    if not newick.endswith(';'):
        raise ValueError("Tree in Newick format must end with ';'")

    root = newick[:-1]
    return parse_newick_subtree(
        root, 
        aggregator, 
        distance_parser=distance_parser, 
        feature_parser=feature_parser
    )
# ---


"""
--------------------------------------------------------------------------------
Usage and tests
--------------------------------------------------------------------------------
To run tests: `pytest parse.py --hypothesis-show-statistics --verbose`
To run static type check: `mypy parse_newick.py`
To run coverage analysis : 
    `coverage run -m pytest parse_newick.py --hypothesis-show-statistics -v`
    `coverage html`
"""

def test_parse_newick():

    def tree_as_list(label, children, distance, features):
        if not children: return label
        return children

    nwk = "((A, B), (C, D));"
    assert (
        parse_newick(nwk, tree_as_list) 
        == 
        [['A','B'], ['C','D']]
    )


    nwk = "(, , );"
    assert (
        parse_newick(nwk, tree_as_list) 
        == 
        ['','','']
    )

    nwk = "((, ), , (, ));"
    assert (
        parse_newick(nwk, tree_as_list) 
        == 
        [['', ''], '', ['', '']]
    )

    def tree_as_dict(label, children, distance, features):
        return dict(
            label=label, 
            children=children, 
            features=(distance, features)
        )

    nwk = 'A;'
    assert (
        parse_newick(nwk, tree_as_dict) 
        == 
        {
            'label': 'A',
            'children': [], 
            'features': (None, '')
        }
    )

    nwk = "(A,);"
    assert (
        parse_newick(nwk, tree_as_dict) 
        == 
        {
            'label': '',
            'children': [
                {'label': 'A', 'children': [], 'features': (None, '')},
                {'label': '', 'children': [], 'features': (None, '')}
            ], 
            'features': (None, '')
        }
    )


    nwk = "(A, B);"
    assert (
        parse_newick(nwk, tree_as_dict) 
        == 
        {
            'label': '', 'features': (None, ''),
            'children': [
                {'label': 'A', 'children': [], 'features': (None, '')},
                {'label': 'B', 'children': [], 'features': (None, '')}
            ]
        }
    )

    nwk = "(A[{'some_feature': 10}], (B:4[some_other_thing], C)D:122.2)root;"
    assert (
        parse_newick(nwk, tree_as_dict) 
        == 
        {
            'label': 'root','features': (None, ''),
            'children': [
                {
                    'label': 'A', 'features': (None, "{'some_feature': 10}"),
                    'children': []
                },
                {
                    'label': 'D', 'features': (122.2, ''),
                    'children': [
                        {
                            'label': 'B', 'features': (4., 'some_other_thing'),
                            'children': []
                        },
                        {
                            'label': 'C', 'features': (None, ''), 'children': []
                        }
                    ]
                }
            ]
        }
    )


    def parse_NHX_features(feature_string):
        '''
        Example: 
          '&&NHX:conf=0.01:name=INTERNAL' -> {'conf':'0.01', 'name': 'INTERNAL'}
        '''
        prefix = '&&NHX:'
        feature_string = feature_string[len(prefix):]
        items = feature_string.split(':')
        key_value_pairs = (item.split('=') for item in items)
        return dict(key_value_pairs)

    def tree_builder(label, children, distance, features):
        return {'label': label, 'children': children, 'features': features}
    
    assert (
        parse_newick(
            'A[&&NHX:conf=0.01:key=value];',
            aggregator=tree_builder,
            feature_parser=parse_NHX_features
        )
        ==
        {'label': 'A', 'children':[], 'features': {'conf':'0.01','key':'value'}}
    )


# <-- Preparing data generators for property testing

from hypothesis.strategies import (text, characters,
                                   recursive, deferred, lists)
from hypothesis.strategies import builds as build_strategy
from string import ascii_letters, digits

Labels = text(ascii_letters + digits + '.-_')
Distances = text(digits + 'e.-')
Comments = text(characters(blacklist_characters='[]()'))

def leaf_builder(label, distance, comment):
    distance_str = (':' + distance) if distance else ''
    comment_str = ('[' + comment + ']') if comment else ''
    return label + distance_str + comment_str

Leafs = build_strategy(leaf_builder, Labels, Distances, Comments)

def children_builder(node_list):
    if node_list:
        return '(' + ','.join(node_list) + ')'

def node_builder(children, label, distance, comment):
    distance_str = (':' + distance) if distance else ''
    comment_str = ('[' + comment + ']') if comment else ''
    return children + label + distance_str + comment_str

NewickNodes = recursive(
    Leafs, 
    lambda children: 
        build_strategy(node_builder, 
            lists(children, min_size=2).map(children_builder), 
            Labels, Distances, Comments),
    max_leaves=500
)

def tree_builder(newick_node):
    return newick_node + ';'

NewickTrees = build_strategy(tree_builder, NewickNodes)


# <-- Property testing

from hypothesis import given, note

def tree_as_dict(label, children, distance, features):
    return {
        'label': label,
        'children': children,
        'distance': distance,
        'features': features
    }
# ---

def dict_tree_to_newick(dict_tree, root_node=False):
    label = dict_tree['label']

    distance = dict_tree['distance']
    distance_str = (':' + distance) if distance else ''

    features = dict_tree['features']
    features_str = ('[' + features + ']') if features else ''

    if dict_tree['children']:
        children = (dict_tree_to_newick(child) for child in dict_tree['children'])
        children_str = '(' + ','.join(children) + ')'
    else:
        children_str = ''

    end = ';' if root_node else ''

    return children_str + label + distance_str + features_str + end


@given(NewickTrees)
def test_newick_decoding_can_be_inverted(newick):

    parsed_tree = parse_newick(newick, tree_as_dict, distance_parser=str)
    note(f'parsed tree: {parsed_tree}')

    assert (
        dict_tree_to_newick(parsed_tree, root_node=True)
        ==
        newick
    )



"""
Generic tree utilities for Python
=================================

Trees are one of the most ubiquitous data structures. It is amazing how often we 
as programmers tend to reimplement the same algorithms for different trees.

This module defines generic tree-traverse and tree-reduce algorithms that can be
used with any tree-like object such as filesystem paths, lists, nested 
dictionaries an expression tree or even specialized tree classes! The only thing 
that must be provided is a function to get child nodes from a parent node.

Examples:

##  Use with any kind of structured tree!
```
    tree = {
        'label':'A', 'children':[
            {'label':'B', 'children':[]},
            {'label':'C', 'children': [
                {'label':'D', 'children':[]}, 
                {'label':'E', 'children':[]}
            ]}
        ]
    }

    def children(node):
        return node['children']
        
    [node['label'] for node in tree_traverse(tree, children, mode='inorder')]

    # Output --> ['B, 'A', 'D', 'C', 'E']
```

##  Even with user-defined classes!
```
    class Tree:
        def __init__(self, label, children=None):
            self.label = label
            self.children = children if children else []

        def is_leaf(self):
            return len(self.children) == 0

    tree = Tree('A', [
            Tree('B'),
            Tree('C',[Tree('D'),Tree('E')])
        ]
    )

    def get_children(node):
        return node.children

    def node_to_newick(node, children):
        if is_leaf(node):
            return node.label
        else:
            return f"({','.join(children)})"


    tree_reduce(tree, get_children, node_to_newick)
                
    # Output --> '(B,(D,E))'
```

## Compose to perform complex algorithms
```
    tree = (('A',('B',('C','D'))),'E')

    def is_leaf(node): 
        return isinstance(node, str)

    def get_children(node): 
        return node if not is_leaf(node) else []

    def induced_subtree(leafs):
        def induced_subtree_generator(node, children):
            if children:
                return tuple(ch for ch in children if not ch is None)
            else:
                return node if node in leafs else None
        return induced_subtree_generator

    leafs = ['B', 'D', 'E']
    induced_subtree = tree_reduce(tree, get_children, induced_subtree(leafs))
    print(induced_subtree)

    # Output --> ((('B',('D',)),),'E')


    def merge_unary_nodes(node, children):
        if is_leaf(node):
            return node
        
        new_children = [
            ch[0] if (len(ch) == 1) else ch
            for ch in children
        ]
        
        return tuple(new_children)

    tree_reduce(induced_subtree, get_children, merge_unary_nodes)

    # Output --> (('B','D'),'E')
```


## Use even with filesystem paths!
```
from pathlib import Path

def enter_folder(path_str):
    return list(Path(path_str).iterdir()) if path.is_dir() else []

for item in tree_traverse('/usr', enter_folder, mode='breadth_first'):
    print(item)

# Output -->
# /
# /proc
# /usr
# ...
# /usr/share
# /usr/bin
# /usr/sbin
# ...
# /usr/bin/jinfo
# /usr/bin/m2400w
# ...
```
"""

from collections import deque
from typing import (Any, List, Callable, Iterable)

TreeLike = Any

def tree_traverse(
        root: TreeLike, 
        get_children: Callable[[TreeLike], List[TreeLike]], 
        mode: str = 'depth_first'
    ) -> Iterable[TreeLike]:
    """
    Iterate over the nodes of the given tree-like object. The descendants of 
    each node are obtained by applying the callable `get_children` on the node. 

    The iteration order is depth-first by default and can be changed via the 
    `mode` argument. The available modes are: 'depth first' (same as 
    'preorder'), 'breadth first' (same as 'level order'), 'inorder' and 
    'postorder', the mode is case insensitive and can contain spaces, 
    underscores and dashes.

    Example:
    ```
    tree = [[['A', 'B'], 'C'], ['D', ['E', 'F'],'G']]
     
    children_fn = (lambda node: node if isinstance(node, list) else [])
    is_leaf = (lambda node: isinstance(node, str))

    # <-- Print leaves in order of traversal
    for node in tree_traverse(tree, children_fn, mode='breadth first') 
        if is_leaf(node):
            print(node, end=' ')
    
    # Prints --> C D G A B E F
    ```
    """

    modes = {
        'depthfirst': traverse_depth_first,
        'preorder': traverse_depth_first,
        'breadthfirst': traverse_breadth_first,
        'levelorder': traverse_breadth_first,
        'postorder': traverse_post_order,
        'inorder': traverse_inorder
    }

    mode = (mode
            .replace('-', '')
            .replace('_', '')
            .replace(' ', '')
            .lower()
        )
        
    if mode not in modes:
        raise ValueError(
            f'Tree traversal mode {mode} not valid, '
            f'valid modes are: {", ".join(modes)}'
        )
    
    traversal_fn = modes[mode]
    yield from traversal_fn(root, get_children)
# ---

def traverse_depth_first(
        tree: TreeLike, 
        get_children: Callable[[TreeLike], List[TreeLike]]
    ) -> Iterable[TreeLike]:

    agenda = [tree]
    while agenda:
        current_node = agenda.pop()
        yield current_node
        agenda.extend(reversed(get_children(current_node)))
# ---

def traverse_breadth_first(
        tree: TreeLike, 
        get_children: Callable[[TreeLike], List[TreeLike]]
    ) -> Iterable[TreeLike]:

    agenda = deque([tree])
    while agenda:
        current_node = agenda.popleft()
        yield current_node
        agenda.extend(get_children(current_node))
# ---

def traverse_post_order(
        tree: TreeLike, 
        get_children: Callable[[TreeLike], List[TreeLike]]
    ) -> Iterable[TreeLike]:

    for child in get_children(tree):
        yield from traverse_post_order(child, get_children)
    yield tree
# ---

def traverse_inorder(
        tree: TreeLike, 
        get_children: Callable[[TreeLike], List[TreeLike]]
    ) -> Iterable[TreeLike]:

    children = get_children(tree)
    if children:
        first, *rest = children

        yield from traverse_inorder(first, get_children)
        yield tree
        for child in rest:
            yield from traverse_inorder(child, get_children)
    else:
        yield tree
# ---

"""
--------------------------------------------------------------------------------
Usage and tests
--------------------------------------------------------------------------------
To run tests: `pytest tree_utils.py`
To run static type check: `mypy tree_utils.py`
To run coverage analysis:
    `coverage run --src=. -m pytest tree_utils.py`
    `coverage html`
"""

def test_traverse():
    tree = [[['A', 'B'], 'C'], ['D', ['E', 'F'],'G']]
     
    children_fn = (lambda node: 
        node if isinstance(node, list) else []
    )

    assert (
        [node for node in tree_traverse(tree, children_fn, mode='depth_first')]
         ==
        [node for node in traverse_depth_first(tree, children_fn)]
    )

    assert (
        [node for node in tree_traverse(tree, children_fn, mode='pre order')]
         ==
        [node for node in traverse_depth_first(tree, children_fn)]
    )

    assert (
        [node for node in tree_traverse(tree, children_fn, mode='breadth-first')]
         ==
        [node for node in traverse_breadth_first(tree, children_fn)]
    )

    assert (
        [node for node in tree_traverse(tree, children_fn, mode='LevelOrder')]
         ==
        [node for node in traverse_breadth_first(tree, children_fn)]
    )

    assert (
        [node for node in tree_traverse(tree, children_fn, mode='POSTorder')]
         ==
        [node for node in traverse_post_order(tree, children_fn)]
    )

    assert (
        [node for node in tree_traverse(tree, children_fn, mode='INORDER')]
         ==
        [node for node in traverse_inorder(tree, children_fn)]
    )

def test_breadth_first():
    tree = [[['A', 'B'], 'C'], ['D', ['E', 'F'],'G']]
     
    children_fn = (lambda node: node if isinstance(node, list) else [])
    is_leaf = (lambda node: isinstance(node, str))

    assert (
        [node 
            for node in traverse_breadth_first(tree, children_fn) 
            if is_leaf(node)]
         ==
         'C,D,G,A,B,E,F'.split(',')
     )

def test_depth_first():
    tree = [[['A', 'B'], 'C'], ['D', ['E', 'F'],'G']]
     
    children_fn = (lambda node: node if isinstance(node, list) else [])
    is_leaf = (lambda node: isinstance(node, str))

    assert (
        [node 
            for node in traverse_depth_first(tree, children_fn) 
            if is_leaf(node)]  
         ==
         'A,B,C,D,E,F,G'.split(',')
     )

def test_post_order():
    tree = [[['A', 'B'], 'C'], ['D', ['E', 'F'],'G']]
     
    children_fn = (lambda node: node if isinstance(node, list) else [])
    is_leaf = (lambda node: isinstance(node, str))

    assert (
        [node 
            for node in traverse_post_order(tree, children_fn) 
            if is_leaf(node)]  
        ==
        'A,B,C,D,E,F,G'.split(',')
    )

    tree = {
        'label':'A', 'children':[
            {'label':'B', 'children':[]},
            {'label':'C', 'children': [
                {'label':'D', 'children':[]}, 
                {'label':'E', 'children':[]}
            ]}
    ]
    }

    def children(node):
        return node['children']
    
    assert (
         [node['label']
            for node in traverse_post_order(tree, children)]
         ==
         'B,D,E,C,A'.split(',')
     )

def test_inorder():
    tree = [[['A', 'B'], 'C'], ['D', ['E', 'F'],'G']]
     
    children_fn = (lambda node: node if isinstance(node, list) else [])
    is_leaf = (lambda node: isinstance(node, str))

    assert (
        [node 
            for node in traverse_post_order(tree, children_fn) 
            if is_leaf(node)]  
        ==
        'A,B,C,D,E,F,G'.split(',')
    )

    tree = {
        'label':'A', 'children':[
            {'label':'B', 'children':[]},
            {'label':'C', 'children': [
                {'label':'D', 'children':[]}, 
                {'label':'E', 'children':[]}
            ]}
    ]
    }

    def children(node):
        return node['children']
    
    assert (
         [node['label']
            for node in traverse_inorder(tree, children)]
         ==
         'B,A,D,C,E'.split(',')
     )

"""
Ideas for property-based testing:
    - Different paths, same destination: Different traversals must yield the 
        same nodes.

    - There, and back again: Converting a tree to another format (lists - dicts)
        and back must yield the same tree.
    
    - Some things never change: Maybe check for invariants?

    - The more things change, the more they stay the same: Applying an operation 
        twice yields the same result as applying it once. Maybe filtering or 
        prunning? Traversing twice?

    - Solve a small problem first: Trees should have recursive properties given 
        their nature.

    - Hard to prove, easy to verify: Maybe yield nodes with paths and then 
        verify that following the path the node is found.
"""


"""
Generic tree utilities for Python
=================================

Trees are one of the most ubiquitous data structures. It is amazing how often we 
as programmers tend to reimplement the same algorithms for different trees.

This module defines generic tree-traverse and tree-reduce algorithms that can be
used with any tree-like object such as filesystem paths, lists, nested 
dictionaries an expression tree or even specialized tree classes! The only thing 
that must be provided is a function to get child nodes from a parent node.

Also, trees are usually represented in some fields (such as bioinformatics) in 
the newick format, which is nontrivial to parse, so this module includes a 
function to do this.

Examples
--------

Import the basic functions:
```
    import treetools
```

###  Use with any kind of structured tree!
```
    tree = {
        'label':'A', 'children':[
            {'label':'B', 'children':[]},
            {'label':'C', 'children': [
                {'label':'D', 'children':[]}, 
                {'label':'E', 'children':[]}
            ]}
        ]
    }

    def children(node):
        return node['children']
        
    [node['label'] 
        for node in treetools.traverse(tree, children, mode='inorder')]

    # Output --> ['B, 'A', 'D', 'C', 'E']
```

###  Even with user-defined classes!
```
    class Tree:
        def __init__(self, label, children=None):
            self.label = label
            self.children = children if children else []

        def is_leaf(self):
            return len(self.children) == 0

    tree = Tree('A', [
            Tree('B'),
            Tree('C',[Tree('D'),Tree('E')])
        ]
    )

    def get_children(node):
        return node.children

    def node_to_newick(node, children):
        if is_leaf(node):
            return node.label
        else:
            return f"({','.join(children)})"


    treetools.reduce(tree, get_children, node_to_newick)
                
    # Output --> '(B,(D,E))'
```
### Parse a newick-formatted tree structure
```
    def parse_data(data_string):
        '''
        Example: 
          'data1=xx,data2=yy' 
            -> {'data1':'xx', 'data2': 'yy'}
        '''
        items = feature_string.split(',')
        key_value_pairs = (item.split('=') for item in items)
        return dict(key_value_pairs)

    def parse_branch_length(length_str):
        return float(length_str) if length_str else 0.0

    def tree_builder(label, children, branch_length, node_data):
        return {
            'label': label, 
            'children': children, 
            'length': branch_length,
            'data': node_data}
    
    newick = "(A:0.2:[dat=23,other=45], B:12.4:[dat=122,other=xyz])root[x=y];"
    
    treetools.parse_newick(
        newick,
        aggregator=tree_builder,
        feature_parser=parse_data,
        distance_parser=parse_branch_length
    )

    # Output ->
    {
        'label': 'root', 'length':0, 'data': {'x':'y'},
        'children': [
            {'label': 'A', 'length':0.2, 'data':{'dat':'23','other':'45'},
             'children': []},
            {'label': 'B', 'length':12.4, 'data':{'dat':'122','other':'xyz'},
             'children': []}, 
        ]
    }
```

### Compose to perform complex algorithms
```
    tree = (('A',('B',('C','D'))),'E')

    def is_leaf(node): 
        return isinstance(node, str)

    def get_children(node): 
        return node if not is_leaf(node) else []

    def induced_subtree(leafs):
        def induced_subtree_generator(node, children):
            if children:
                return tuple(ch for ch in children if not ch is None)
            else:
                return node if node in leafs else None
        return induced_subtree_generator

    leafs = ['B', 'D', 'E']
    induced = treetools.reduce(tree, get_children, induced_subtree(leafs))
    print(induced)

    # Output --> ((('B',('D',)),),'E')


    def merge_unary_nodes(node, children):
        if is_leaf(node):
            return node
        
        new_children = [
            ch[0] if (len(ch) == 1) else ch
            for ch in children
        ]
        
        return tuple(new_children)

    treetools.reduce(subtree, get_children, merge_unary_nodes)

    # Output --> (('B','D'),'E')
```


### Use even with filesystem paths!
```
from pathlib import Path

def enter_folder(path_str):
    return list(Path(path_str).iterdir()) if path.is_dir() else []

for item in treetools.traverse('/usr', enter_folder, mode='breadth_first'):
    print(item)

# Output -->
# /
# /proc
# /usr
# ...
# /usr/share
# /usr/bin
# /usr/sbin
# ...
# /usr/bin/jinfo
# /usr/bin/m2400w
# ...
```
"""

from .traverse import tree_traverse as traverse
from .reduce import tree_reduce as reduce
from .parse import parse_newick, parse_newick_subtree


"""The setup script for packaging.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get the version from the VERSION file
with open(path.join(here, 'VERSION')) as version_file:
    version = version_file.read().strip()

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project:
    #
    # $ pip install treet
    #
    # And where it will live on PyPI: https://pypi.org/project/treet/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='treet',  # Required

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=version,  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description=('Simple, minimal, but powerful tools to handle any kind of '
                'hierarchical (tree) structures'),  # Required

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional
    long_description_content_type="text/markdown",

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='https://github.com/Ad115/python-treet',  # Optional

    # This should be your name or the name of the organization which owns the
    # project.
    author='Ad115',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='a.garcia230395@gmail.com',  # Optional

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Intended Audience :: Developers',

        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Scientific/Engineering :: Artificial Life',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Text Processing',
        'Topic :: Utilities',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3 :: Only',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='tree hierarchical generic newick traverse reduce',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[],

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={  # Optional
        'dev': ['pytest', 'hypothesis', 'coverage', 'mypy']
    },

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    project_urls={  # Optional
        'Say Thanks!': 'https://saythanks.io/to/Ad115',
        'Documentation': 'https://github.com/Ad115/python-treet',
        'Author': 'https://agargar.wordpress.com/'
    }
)
